from distutils.command.build_ext import build_ext
from distutils.core import Distribution
from distutils.errors import (CCompilerError, CompileError, DistutilsExecError,
                              DistutilsPlatformError)
from distutils.extension import Extension
from os import chmod, stat
from pathlib import Path
from shutil import copyfile
from subprocess import run


class GoExtension(Extension):
    def __init__(self, name, path):
        super().__init__(name, sources=[])
        self.path = path


extensions = [
    GoExtension(
        #"groove",
        "groove.assets.grooveproxy",
        # Assume we have temporarily copied over the proxy folder into our current path
        # We don't want it to be referenced in the actual parent library, since we need to bundle
        # it alongside the python project in sdist in case clients need to build from source
        # when wheels aren't available.
        "./proxy",
    )
]


class BuildFailed(Exception):
    pass


class GoExtensionBuilder(build_ext):
    def run(self):
        try:
            build_ext.run(self)
        except (DistutilsPlatformError, FileNotFoundError):
            raise BuildFailed("File not found. Could not compile extension.")

    def build_extension(self, ext):
        try:
            if isinstance(ext, GoExtension):
                extension_root = Path(__file__).parent.resolve() / ext.path
                ext_path = self.get_ext_fullpath(ext.name)
                result = run(["go", "build", "-o", str(Path(ext_path).absolute())], cwd=extension_root)
                if result.returncode != 0:
                    raise CompileError("Go build failed")
            else:
                build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError, ValueError):
            raise BuildFailed('Could not compile C extension.')


def build(setup_kwargs):
    distribution = Distribution({"name": "python_ctypes", "ext_modules": extensions})
    distribution.package_dir = "python_ctypes"

    cmd = GoExtensionBuilder(distribution)
    cmd.ensure_finalized()
    cmd.run()

    # This is somewhat of a hack with go executables; this pipeline will package
    # them as .so files but they aren't actually built libraries. We maintain
    # this convention only for the ease of plugging in to poetry and distutils that
    # use this suffix to indicate the build architecture and run on the
    # correct downstream client OS.
    for output in cmd.get_outputs():
        relative_extension = Path(output).relative_to(cmd.build_lib)
        copyfile(output, relative_extension)
        mode = stat(relative_extension).st_mode
        mode |= (mode & 0o444) >> 2
        chmod(relative_extension, mode)


if __name__ == "__main__":
    build({})



from contextlib import contextmanager
from enum import Enum
from subprocess import Popen
from sysconfig import get_config_var
from time import sleep
from urllib.parse import urljoin

from requests import Session

from groove.assets import get_asset_path
from groove.dialer import DefaultInternetDialer, DialerDefinition
from groove.enums import CacheModeEnum
from groove.tape import TapeSession


class ProxyFailureError(Exception):
    pass


class Groove:
    def __init__(
        self,
        command_timeout: int = 5,
        port: int = 6010,
        control_port: int = 6011,
        auth_username: str | None = None,
        auth_password: str | None = None,
    ):
        self.session = Session()

        self.port = port
        self.control_port = control_port
        self.auth_username = auth_username
        self.auth_password = auth_password

        self.base_url_proxy = f"http://localhost:{port}"
        self.base_url_control = f"http://localhost:{control_port}"

        self.timeout = command_timeout

    @contextmanager
    def launch(self):
        parameters = {
            "--port": self.port,
            "--control-port": self.control_port,
            "--auth-username": self.auth_username,
            "--auth-password": self.auth_password,
        }

        # Not specifying parameters should make them null
        parameters = {key: value for key, value in parameters.items() if value is not None}

        process = Popen(
            [
                self.executable_path,
                *[
                    str(item)
                    for key, value in parameters.items()
                    for item in [key, value]
                ]
            ]
        )

        # Wait for launch
        sleep(0.5)

        try:
            yield
        finally:
            process.terminate()

    def tape_start(self):
        response = self.session.post(urljoin(self.base_url_control, "/api/tape/record"), timeout=self.timeout)
        assert response.json()["success"] == True

    def tape_get(self, tape_id: str | None = None) -> TapeSession:
        tape_response = self.session.post(
            urljoin(self.base_url_control, "/api/tape/retrieve"),
            json=dict(
                tapeID=tape_id,
            ),
            timeout=self.timeout
        )
        return TapeSession.from_server(tape_response.content)

    def tape_load(self, session: TapeSession):
        tape_response = self.session.post(
            urljoin(self.base_url_control, "/api/tape/load"),
            files={"file": session.to_server()},
            timeout=self.timeout,
        )
        assert tape_response.json()["success"] == True

    def tape_stop(self):
        response = self.session.post(urljoin(self.base_url_control, "/api/tape/stop"), timeout=self.timeout)
        assert response.json()["success"] == True

    def tape_clear(self, tape_id: str | None = None):
        response = self.session.post(
            urljoin(self.base_url_control, "/api/tape/clear"),
            json=dict(
                tapeID=tape_id,
            ),
            timeout=self.timeout,
        )
        assert response.json()["success"] == True

    def set_cache_mode(self, mode: CacheModeEnum):
        response = self.session.post(
            urljoin(self.base_url_control, "/api/cache/mode"),
            json=dict(
                mode=mode.value,
            )
        )
        assert response.json()["success"] == True

    def cache_clear(self):
        response = self.session.post(urljoin(self.base_url_control, "/api/cache/clear"), timeout=self.timeout)
        assert response.json()["success"] == True

    def dialer_load(self, dialers: list[DialerDefinition]):
        response = self.session.post(
            urljoin(self.base_url_control, "/api/dialer/load"),
            json=dict(
                definitions=[
                    {
                        "priority": dialer.priority,
                        "proxyServer": dialer.proxy.url if dialer.proxy is not None else None,
                        "proxyUsername": dialer.proxy.username if dialer.proxy is not None else None,
                        "proxyPassword": dialer.proxy.password if dialer.proxy is not None else None,
                        "requiresUrlRegex": dialer.request_requires.url_regex if dialer.request_requires is not None else None,
                        "requiresResourceTypes": dialer.request_requires.resource_types if dialer.request_requires is not None else None,
                    }
                    for dialer in dialers
                ],
            )
        )
        assert response.json()["success"] == True

    @property
    def executable_path(self) -> str:
        # Support statically and dynamically build libraries
        if (path := get_asset_path("grooveproxy")).exists():
            return str(path)

        wheel_extension = get_config_var("EXT_SUFFIX")
        if (path := get_asset_path(f"grooveproxy{wheel_extension}")).exists():
            return exit(path)

        raise ValueError("No groove executable file found")



from base64 import b64decode, b64encode
from gzip import compress, decompress
from json import dumps, loads

from pydantic import validator

from groove.models import GrooveModelBase


class TapeRequest(GrooveModelBase):
    url: str
    method: str
    headers: dict[str, list[str]]
    body: bytes

    @validator("body")
    def validate_body(cls, value):
        return b64decode(value)


class TapeResponse(GrooveModelBase):
    status: int
    headers: dict[str, list[str]]
    body: bytes

    @validator("body")
    def validate_body(cls, value):
        return b64decode(value)


class TapeRecord(GrooveModelBase):
    request: TapeRequest
    response: TapeResponse

    class Config:
        json_encoders = {
            # Assume that body bytes should always be encoded as base64 strings
            bytes: lambda value: b64encode(value).decode(),
        }


class TapeSession(GrooveModelBase):
    records: list[TapeRecord]

    @classmethod
    def from_server(cls, data: bytes):
        raw_records = loads(decompress(data))
        return cls(records=raw_records)

    def to_server(self) -> bytes:
        return compress(
            dumps([
                # json_encoders doesn't operate on list items, must iterate manually
                # https://github.com/pydantic/pydantic/issues/4085
                loads(record.json())
                for record in self.records
            ]).encode()
        )


import pytest
from playwright.sync_api import sync_playwright

from groove.proxy import Groove
from requests import Session


# We want this to recreate by default on every unit test to clear the state
@pytest.fixture(scope="function")
def proxy():
    proxy = Groove()
    with proxy.launch():
        # Before we yield to new client calls clear out any remaining cache items
        proxy.cache_clear()
        yield proxy

@pytest.fixture(scope="function")
def browser():
    with sync_playwright() as p:
        yield p.chromium.launch(
            headless=True,
        )

@pytest.fixture
def context(proxy, browser):
    yield browser.new_context(
        proxy={
            "server": proxy.base_url_proxy,
        }
    )

@pytest.fixture(scope="function")
def session(proxy):
    session = Session()
    session.proxies = {
        "http": proxy.base_url_proxy,
        "https": proxy.base_url_proxy,
    }
    yield session



from base64 import b64encode
from uuid import uuid4

from bs4 import BeautifulSoup
from functools import partial
from requests import get

from groove.tape import TapeRecord, TapeRequest, TapeResponse, TapeSession
from groove.tests.mock_server import MockPageDefinition, mock_server


def test_tape_global(proxy, browser):
    """
    Ensure the basic tape functions work correctly
    """
    proxy.tape_start()

    # Explicitly use different contexts because Chromium will cache this page client side
    context = browser.new_context(
        proxy={
            "server": proxy.base_url_proxy,
        }
    )
    page = context.new_page()
    page.goto("https://freeman.vc")
    page.close()

    modified_records = 0
    session = proxy.tape_get()
    assert len(session.records) > 0

    for record in session.records:
        if record.request.url == "https://freeman.vc:443/":
            record.response.body = "Mocked content".encode()
            modified_records += 1
    assert modified_records == 1

    proxy.tape_load(session)

    context = browser.new_context(
        proxy={
            "server": proxy.base_url_proxy,
        }
    )
    page = context.new_page()
    page.goto("https://freeman.vc")

    assert BeautifulSoup(page.content(), features="html.parser").text.strip() == "Mocked content"


def test_tape_id(proxy, session):
    """
    Ensure tapes can be recorded separately
    """
    proxy.tape_start()

    with mock_server([
        MockPageDefinition(
            "/test1",
            content=f"<html><body>Request 1</body></html>"
        ),
        MockPageDefinition(
            "/test2",
            content=f"<html><body>Request 2</body></html>"
        ),
    ]) as mock_url:
        response1 = session.get(f"{mock_url}/test1", headers={"Tape-ID": "Tape1"})
        assert response1.ok
        response2 = session.get(f"{mock_url}/test2", headers={"Tape-ID": "Tape2"})
        assert response2.ok

    session1 = proxy.tape_get("Tape1")
    session2 = proxy.tape_get("Tape2")
    assert len(session1.records) == 1
    assert len(session2.records) == 1

    assert session1.records[0].request.url == f"{mock_url}/test1"
    assert session2.records[0].request.url == f"{mock_url}/test2"


def test_multiple_requests(proxy, context):
    """
    Ensure mocked requests resolve in the same order
    """
    response_1 = str(uuid4())
    response_2 = str(uuid4())

    records = [
        TapeRecord(
            request=TapeRequest(
                url="https://freeman.vc:443/",
                method="GET",
                headers={},
                body=b"",
            ),
            response=TapeResponse(
                status=200,
                headers={},
                body=b64encode(response.encode())
            ),
        )
        for response in [response_1, response_2]
    ]

    proxy.tape_load(
        TapeSession(
            records=records
        )
    )

    page = context.new_page()

    page.goto("https://freeman.vc")
    assert BeautifulSoup(page.content(), features="html.parser").text.strip() == response_1

    page.goto("https://freeman.vc")
    assert BeautifulSoup(page.content(), features="html.parser").text.strip() == response_2



import pytest
from playwright._impl._api_types import Error as PlaywrightError

from groove.proxy import ProxyFailureError


@pytest.mark.xfail(reason="TLS certificate blocked", raises=ProxyFailureError)
def test_tls_addons(proxy, context):
    """
    Test TLS addons
    """
    page = context.new_page()

    try:
        page.goto("https://www.google.com:443/")
    except PlaywrightError as e:
        if "net::ERR_EMPTY_RESPONSE" in e.message:
            raise ProxyFailureError()



from groove.models import GrooveModelBase


class RequestRequiresDefinition(GrooveModelBase):
    url_regex: str
    resource_types: list[str]


class ProxyDefinition(GrooveModelBase):
    url: str
    username: str | None = None
    password: str | None = None


class DialerDefinition(GrooveModelBase):
    priority: int
    proxy: ProxyDefinition | None = None
    request_requires: RequestRequiresDefinition | None = None


class DefaultInternetDialer(DialerDefinition):
    """
    Proxy all requests to the open internet, with low priority
    """
    def __init__(self):
        super().__init__(
            priority=1,
            proxy=None,
            request_requires=None,
        )

class DefaultLocalPassthroughDialer(DialerDefinition):
    """
    Proxy generally static assets to the open internet, with high priority
    """
    def __init__(self):
        super().__init__(
            priority=1000,
            proxy=None,
            request_requires=RequestRequiresDefinition(
                url_regex=".*?.(?:txt|json|css|less|js|mjs|cjs|gif|ico|jpe?g|svg|png|webp|mkv|mp4|mpe?g|webm|eot|ttf|woff2?)",
                resource_types=["script", "image", "stylesheet", "media", "font"],
            ),
        )
